#ifndef H_DESSIN
#define H_DESSIN

// On dfinit la classe DESSIN qui contient
// - TOPOLOGIE : la topologie, qui reste inchange
// - VECTEUR : la gomtrie (position des points)
// - BELYI : la gomtrie (fonction de Belyi)
// - les valeurs temporaires pour la rsolution de l'quation f(x) = 0 :
//   oppos de l'image  - f(x)  et drive  f'(x)
// - les ancres (quations supplmentaires)

#include "topologie.h"
#include "belyi.h"

enum exceptions { invalid, lecture };

class SELECTION {
  ulong longueur;
  ulong* liste;
  enum { impossible, absent, present };
public:
  ~SELECTION () { delete[] liste; }
  SELECTION (ulong nbpt, bool possible);
  SELECTION (const SELECTION&);
  SELECTION& operator = (const SELECTION&);
  bool defini () const { return longueur > 0; }
  ulong nombre () const;
  bool is_present (ulong i) const;
  bool is_possible (ulong i) const;
  bool same_possible (const SELECTION&) const;
  void set_impossible (ulong i);
  void set_possible (ulong i);
  void swap_possible (ulong i);
};

class ANCRE {
  enum { chunk = 10 };
  ulong ancre_nombre;
  ulong* ancre_liste;
public:
  enum { nondefini, point, somme, moyenne, produit } type;
  ~ANCRE () { delete[] ancre_liste; }
  ANCRE () : ancre_nombre(0), ancre_liste(new ulong[chunk]), type(nondefini) {}
  ANCRE (ulong pt) : ancre_nombre(1), ancre_liste(new ulong[chunk]), type(point)
    { ancre_liste[0] = pt; }
  ANCRE (const ANCRE&);

  ulong nombre () const { return ancre_nombre; }
  ulong operator [] (const ulong i) const
    { return (i<ancre_nombre) ? ancre_liste[i] : 0; }
  void add_or_delete (const ulong pt);
  void fusion (ulong i, ulong j, ulong max);
  void split (ulong newpoint);
private: // On supprime les oprateurs par dfaut
  ANCRE& operator = (const ANCRE&);
};
extern CX negvaleur (const ANCRE& a, const VECTEUR& x);
extern CX derivee   (const ANCRE& a, const VECTEUR& x, ulong i);


class DESSIN
{
  // Donnes du dessin (pointeurs)
  TOPOLOGIE* dessin_topo;
  VECTEUR*   dessin_pos;
  BELYI*     dessin_belyi;
  // Primages de la droite relle
  public:
  struct LIEN { ulong i, j; LIEN() : i(0), j(0) {} }; // Il faut i > j
  struct LIENS {
    ulong nb; LIEN* val;
    ~LIENS() { delete val; }
    LIENS(const TOPOLOGIE& t) {
      nb = t.degre(); if (t.nombrevaleurs() == 4) nb += t.degre();
      val = new LIEN[nb]; }
    LIENS(ulong s) : nb(s), val(new LIEN[s]) { }
    LIEN& operator[] (ulong n)
      { if (n<nb) return val[n]; else throw *this; }
    const LIEN& operator[] (ulong n) const
      { if (n<nb) return val[n]; else throw *this; }
  private: // On supprime les oprateurs par dfaut
    LIENS& operator = (const LIENS&);
    LIENS (const LIENS&);
  };
  private:
  LIENS* dessin_liens;
  bool liens_fixes;
  VECTEUR* liens0p; // position des primages de 0
  VECTEUR* liens0q; // position des primages de 1
  VECTEUR* liens1; // position des primages de 1/2
  VECTEUR* liens2p; // position des primages de 1/4
  VECTEUR* liens2q; // position des primages de 3/4
  VECTEUR* liens3p; // position des primages de 1/8
  VECTEUR* liens3q; // position des primages de 7/8
  VECTEUR* liens5p; // position des primages de  1/32
  VECTEUR* liens5q; // position des primages de 31/32
  VECTEUR* liens8p; // position des primages de   1/256
  VECTEUR* liens8q; // position des primages de 255/256
  VECTEUR* liens13p; // position des primages de    1/8192
  VECTEUR* liens13q; // position des primages de 8191/8192
  ulong liens_morceaux;
  // Donnes du systme  rsoudre
  ANCRE *ancre0, *ancre1;
  VECTEUR* dessin_negimage;
  MCARREE* dessin_derivee;

  const TOPOLOGIE& topologie () const { return *dessin_topo; }
  const BELYI& belyi ();
  void calcule_ancres ();
  void fixeliens ();
  void demiliens ();
  void calcule_liens ();
  void demi_positions ();
  VECTEUR negimage (const BELYI& b, const VECTEUR& x) const;
  VECTEUR derivee_fast (ulong i) const;
  VECTEUR derivee_tangente (ulong i) const;
  ulong voisin (const CX& c) const;
  void info (const CX& c) const;
  void moved ();
  void empile ()
    { historique.empile (topologie(), position(), *ancre0, *ancre1); }
  void depile ()
    { historique.depile (dessin_topo, dessin_pos,  ancre0,  ancre1); }
  // direction pour le split
  VECTEUR* liste_voisins (const ulong mark);
  SELECTION voisins (const ulong mark);
  // Historique
  class HISTORIQUE {
    struct MEMOIRE {
      TOPOLOGIE *t;
      VECTEUR   *x;
      ANCRE *a0, *a1;
      ~MEMOIRE ();
      MEMOIRE ();
    private: // On supprime les oprateurs par dfaut
      MEMOIRE& operator = (const MEMOIRE&);
      MEMOIRE (const MEMOIRE&);
    } * hist_archives;
    ulong hist_maxdepth;
    ulong hist_depth;
    void depile (TOPOLOGIE*& t, VECTEUR*& x, ANCRE*& a0, ANCRE*& a1);
    void empile (const TOPOLOGIE& t, const VECTEUR& x,
                 const ANCRE& a0, const ANCRE& a1);
  public:
    HISTORIQUE (ulong d);
    ~HISTORIQUE ();
    friend void DESSIN::empile ();
    friend void DESSIN::depile ();
  private: // On supprime les oprateurs par dfaut
    HISTORIQUE& operator = (const HISTORIQUE&);
    HISTORIQUE (const HISTORIQUE&);
  };
  HISTORIQUE historique;
public:
  DESSIN (const char* filename);
  ~DESSIN ();

  const RE& C() const { return topologie().C(); }
  const ulong nombrevaleurs () const { return topologie().nombrevaleurs(); }
  const VECTEUR& position () const { return *dessin_pos; }
  const VECTEUR& negimage ();
  const MCARREE& derivee ();
  void undo ();
  void move (const VECTEUR& decalage);
  void moveC (const RE& newC);
  void move_ancre (const ulong mark, ANCRE& a);
  void ajustelambda (bool empiler = true);
  void ajusteancres ();
  void unsplit ();
  void split (ulong mark, SELECTION selection);
  void inversion (const ulong mark);
  void precision (bool augmente);
  void chargement (const char* filename);
  friend ostream& operator << (ostream& s, const DESSIN& d);

  friend class INTERACTIF;
  friend VECTEUR negimage_currentDESSIN (const VECTEUR& x);
private: // On supprime les oprateurs par dfaut
  DESSIN& operator = (const DESSIN&);
  DESSIN (const DESSIN&);
};
extern void set_currentDESSIN(const DESSIN* d);

#endif
