#include "interactif.h"
#include "solve.h"
#include "algdep.h"
#include <stdlib.h> // rand48

static void
dessine_ancre(const ANCRE&a, const COULEUR color, const ulong textnum,
  const TOPOLOGIE&t, const VECTEUR&x, const FENETRE&f)
{
  for (ulong j = 0; j < a.nombre(); j++) f.cercle(8, x[a[j]], 0, color);
  char* texte = "";
  switch (a.type) {
    case ANCRE::nondefini: break;
    case ANCRE::point:   texte = "point";   break;
    case ANCRE::somme:   texte = "somme";   break;
    case ANCRE::moyenne: texte = "moyenne"; break;
    case ANCRE::produit: texte = "produit"; break;
  }
  f.textaffiche(textnum, texte, -60, (textnum==6)?-40:-10, color);
}

static COULEUR couleur_X1 = 0;
static COULEUR couleur_X2 = 0;
static COULEUR couleur_X3 = 0;
static COULEUR couleur_X5 = 0;
static COULEUR couleur_X8 = 0;
static COULEUR couleur_X13 = 0;
INTERACTIF::INTERACTIF (const char* winname, DESSIN& d)
: FENETRE (winname)
, dessin (&d), mark (0)
, mark_voisins (SELECTION(0,false))
, couleur_face   (RGBcolor(0.,1.,0.))
, couleur_ancre0 (RGBcolor(0.,0.,1.))
, couleur_ancre1 (RGBcolor(0.,1.,1.))
, couleur_marque (RGBcolor(1.,0.,0.))
{
  affiche();
  couleur_X1 = RGBcolor(1.,0.,.2);
  couleur_X2 = RGBcolor(1.,0.,.5);
  couleur_X3 = RGBcolor(1.,0.,1.);
  couleur_X5 = RGBcolor(.5,0.,1.);
  couleur_X8 = RGBcolor(.2,0.,1.);
  couleur_X13 = RGBcolor(.1,0.,1.);
}

void
INTERACTIF::affiche (const ulong f, const bool sinon_efface)
{
#if defined(mplib_pari)
  cout << "-- (avant affiche)\n";
  cout << "   Mmoire utilise   : " << top-avma
       << " (" << 100.*(top-avma)/(top-bot) << "%)\n";
#endif
  const VECTEUR& x = dessin->position();
  const TOPOLOGIE& t = dessin->topologie();
  const ulong nbf = t.faces().n();
  const ulong nbs = t.sommets().n();
  const ulong nba = t.aretes().n();
  const ulong nbq = t.quatriemes().n();
  if (nbf + nbs + nba + nbq != x.length()) {
    cerr << "Erreur (affiche) : " << x.length() << " variables, mais "
         << (nbf + nbs + nba + nbq) << " points  afficher\n";
    return;
  }
  if (f & traceaxes) axes();
  const COULEUR c_d = !sinon_efface ? couleur_fond : couleur_dessin;
  const COULEUR c_f = !sinon_efface ? couleur_fond : couleur_face;
  const COULEUR c_0 = !sinon_efface ? couleur_fond : couleur_ancre0;
  const COULEUR c_1 = !sinon_efface ? couleur_fond : couleur_ancre1;
  const COULEUR c_m = !sinon_efface ? couleur_fond : couleur_marque;
  const COULEUR c_X1 = !sinon_efface ? couleur_fond : couleur_X1;
  const COULEUR c_X2 = !sinon_efface ? couleur_fond : couleur_X2;
  const COULEUR c_X3 = !sinon_efface ? couleur_fond : couleur_X3;
  const COULEUR c_X5 = !sinon_efface ? couleur_fond : couleur_X5;
  const COULEUR c_X8 = !sinon_efface ? couleur_fond : couleur_X8;
  const COULEUR c_X13= !sinon_efface ? couleur_fond : couleur_X13;
  if (f & legende) {
    textaffiche(0,"Faces",  -130,-10,c_d); cercle(4,-140,-15, true, c_f);
    textaffiche(1,"Sommets",-130,-25,c_d); cercle(3,-140,-30, true, c_d);
    textaffiche(2,"Artes", -130,-40,c_d); croix (5,-140,-45, c_d);
    if (nbq) { textaffiche(3, "Quatrimes", -130,-55, c_d);
               cercle(3, -140,-60, false, c_d); }
    textaffiche(4,"Ancre 0", -60, -55, c_0);
    textaffiche(5,"Ancre 1", -60, -25, c_1);
    char texte[100]; sprintf(texte, "Degr : %ld", t.degre());
    textaffiche(14, texte, -110, 30, c_d);
  }
  if (f & points) {
    {for (ulong i=1; i<nbf; i++) cercle(4, x[i],     true, c_f);}
    {for (ulong i=0; i<nbs; i++) cercle(3, x[i+nbf], true, c_d);}
    {for (ulong i=0; i<nba; i++) croix (5, x[i+nbf+nbs], c_d);}
    {for (ulong i=0; i<nbq; i++) cercle(3, x[i+nbf+nbs+nba], false, c_d);}
  }
  if (f & ancres) {
    dessine_ancre(*dessin->ancre0, c_0, 6, t, x, *this);
    dessine_ancre(*dessin->ancre1, c_1, 7, t, x, *this);
  }
  if (f & liens) {
    char texte[100];
    switch (dessin->liens_morceaux) {
    case 0:sprintf(texte,"%s",dessin->liens_fixes?"liens fixes":"");break;
    case 1:sprintf(texte,"demi-liens%s",dessin->liens_fixes?" fixes":"");break;
    case 2:sprintf(texte,"quart-liens%s",dessin->liens_fixes?" fixes":"");break;
    case 3:sprintf(texte,"oct-liens%s",dessin->liens_fixes?" fixes":"");break;
    case 5:sprintf(texte,"1/32-liens%s",dessin->liens_fixes?" fixes":"");break;
    case 8:sprintf(texte,"1/256-liens%s",dessin->liens_fixes?" fixes":"");break;
    case 13:sprintf(texte,"1/8192-liens%s",dessin->liens_fixes?" fixes":"");break;
    default:sprintf(texte,"X-liens%s",dessin->liens_fixes?" fixes":"");break;
    }
    textaffiche(13, texte, -110, 50, c_d);
    dessin->calcule_liens();
    DESSIN::LIENS& lnks = *(dessin->dessin_liens);
    if (dessin->liens_morceaux>0) {
      dessin->demi_positions ();
      const VECTEUR vp = *dessin->liens0p;
      const VECTEUR vq = *dessin->liens0q;
      switch (dessin->liens_morceaux) {
        case 1:
          for (ulong i=0; i<lnks.nb; i++) {
            ligne(vp[i], (*dessin->liens1)[i], c_d);
            ligne(vq[i], (*dessin->liens1)[i], c_d);
            cercle(4, (*dessin->liens1 )[i], false, c_X1);
          }
          break;
        case 2:
          for (ulong i=0; i<lnks.nb; i++) {
            ligne(vp[i], (*dessin->liens2p)[i], c_d);
            ligne((*dessin->liens1)[i], (*dessin->liens2p)[i], c_d);
            ligne((*dessin->liens1)[i], (*dessin->liens2q)[i], c_d);
            ligne(vq[i], (*dessin->liens2q)[i], c_d);
            cercle(4, (*dessin->liens2p)[i], false, c_X2);
            cercle(4, (*dessin->liens2q)[i], false, c_X2);
            cercle(4, (*dessin->liens1 )[i], false, c_X1);
          }
          break;
        case 3:
          for (ulong i=0; i<lnks.nb; i++) {
            ligne(vp[i], (*dessin->liens3p)[i], c_d);
            ligne((*dessin->liens3p)[i], (*dessin->liens2p)[i], c_d);
            ligne((*dessin->liens1)[i],  (*dessin->liens2p)[i], c_d);
            ligne((*dessin->liens1)[i],  (*dessin->liens2q)[i], c_d);
            ligne((*dessin->liens3q)[i], (*dessin->liens2q)[i], c_d);
            ligne(vq[i], (*dessin->liens3q)[i], c_d);
            cercle(4, (*dessin->liens3p)[i], false, c_X3);
            cercle(4, (*dessin->liens3q)[i], false, c_X3);
            cercle(4, (*dessin->liens2p)[i], false, c_X2);
            cercle(4, (*dessin->liens2q)[i], false, c_X2);
            cercle(4, (*dessin->liens1 )[i], false, c_X1);
          }
          break;
        case 5:
          for (ulong i=0; i<lnks.nb; i++) {
            ligne(vp[i], (*dessin->liens5p)[i], c_d);
            ligne((*dessin->liens3p)[i], (*dessin->liens5p)[i], c_d);
            ligne((*dessin->liens3p)[i], (*dessin->liens2p)[i], c_d);
            ligne((*dessin->liens1)[i],  (*dessin->liens2p)[i], c_d);
            ligne((*dessin->liens1)[i],  (*dessin->liens2q)[i], c_d);
            ligne((*dessin->liens3q)[i], (*dessin->liens2q)[i], c_d);
            ligne((*dessin->liens3q)[i], (*dessin->liens5q)[i], c_d);
            ligne(vq[i], (*dessin->liens5q)[i], c_d);
            cercle(4, (*dessin->liens5p)[i], false, c_X5);
            cercle(4, (*dessin->liens5q)[i], false, c_X5);
            cercle(4, (*dessin->liens3p)[i], false, c_X3);
            cercle(4, (*dessin->liens3q)[i], false, c_X3);
            cercle(4, (*dessin->liens2p)[i], false, c_X2);
            cercle(4, (*dessin->liens2q)[i], false, c_X2);
            cercle(4, (*dessin->liens1 )[i], false, c_X1);
          }
          break;
        case 8:
          for (ulong i=0; i<lnks.nb; i++) {
            ligne(vp[i], (*dessin->liens8p)[i], c_d);
            ligne((*dessin->liens8p)[i], (*dessin->liens5p)[i], c_d);
            ligne((*dessin->liens3p)[i], (*dessin->liens5p)[i], c_d);
            ligne((*dessin->liens3p)[i], (*dessin->liens2p)[i], c_d);
            ligne((*dessin->liens1)[i],  (*dessin->liens2p)[i], c_d);
            ligne((*dessin->liens1)[i],  (*dessin->liens2q)[i], c_d);
            ligne((*dessin->liens3q)[i], (*dessin->liens2q)[i], c_d);
            ligne((*dessin->liens3q)[i], (*dessin->liens5q)[i], c_d);
            ligne((*dessin->liens8q)[i], (*dessin->liens5q)[i], c_d);
            ligne(vq[i], (*dessin->liens8q)[i], c_d);
            cercle(4, (*dessin->liens8p)[i], false, c_X8);
            cercle(4, (*dessin->liens8q)[i], false, c_X8);
            cercle(4, (*dessin->liens5p)[i], false, c_X5);
            cercle(4, (*dessin->liens5q)[i], false, c_X5);
            cercle(4, (*dessin->liens3p)[i], false, c_X3);
            cercle(4, (*dessin->liens3q)[i], false, c_X3);
            cercle(4, (*dessin->liens2p)[i], false, c_X2);
            cercle(4, (*dessin->liens2q)[i], false, c_X2);
            cercle(4, (*dessin->liens1 )[i], false, c_X1);
          }
          break;
        case 13:
          for (ulong i=0; i<lnks.nb; i++) {
            ligne(vp[i], (*dessin->liens13p)[i], c_d);
            ligne((*dessin->liens8p)[i], (*dessin->liens13p)[i], c_d);
            ligne((*dessin->liens8p)[i], (*dessin->liens5p)[i], c_d);
            ligne((*dessin->liens3p)[i], (*dessin->liens5p)[i], c_d);
            ligne((*dessin->liens3p)[i], (*dessin->liens2p)[i], c_d);
            ligne((*dessin->liens1)[i],  (*dessin->liens2p)[i], c_d);
            ligne((*dessin->liens1)[i],  (*dessin->liens2q)[i], c_d);
            ligne((*dessin->liens3q)[i], (*dessin->liens2q)[i], c_d);
            ligne((*dessin->liens3q)[i], (*dessin->liens5q)[i], c_d);
            ligne((*dessin->liens8q)[i], (*dessin->liens5q)[i], c_d);
            ligne((*dessin->liens8q)[i], (*dessin->liens13q)[i], c_d);
            ligne(vq[i], (*dessin->liens13q)[i], c_d);
            cercle(4, (*dessin->liens13p)[i], false, c_X13);
            cercle(4, (*dessin->liens13q)[i], false, c_X13);
            cercle(4, (*dessin->liens8p)[i], false, c_X8);
            cercle(4, (*dessin->liens8q)[i], false, c_X8);
            cercle(4, (*dessin->liens5p)[i], false, c_X5);
            cercle(4, (*dessin->liens5q)[i], false, c_X5);
            cercle(4, (*dessin->liens3p)[i], false, c_X3);
            cercle(4, (*dessin->liens3q)[i], false, c_X3);
            cercle(4, (*dessin->liens2p)[i], false, c_X2);
            cercle(4, (*dessin->liens2q)[i], false, c_X2);
            cercle(4, (*dessin->liens1 )[i], false, c_X1);
          }
          break;
        default:
          cerr << "Never Reached\n"; throw this;
      }
    }
    else
      for (ulong i=0; i<x.length()-2; i++)
        ligne(x[lnks[i].i], x[lnks[i].j], c_d);
  }
  if (f & etoile) {
    // TODO : liens vers les centres de faces
    // pointilles (x[lnks[k].i], x[lnks[k].j], c_d);
  }
  if (f & marque) {
    if (mark) cercle(10, x[mark], 0, c_m);
  }
  if (f & voisins) if (mark_voisins.defini()) {
    dessin->calcule_liens ();
    dessin->demi_positions ();
    const CX& pos_marque = x[mark];
    VECTEUR* pos_fin = dessin->liste_voisins (mark);
    for (ulong k = 0; k < dessin->topologie().degre(); k++)
      if (mark_voisins.is_possible (k))
        ligne (pos_marque, (*pos_fin)[k],
          mark_voisins.is_present (k) ? c_m : c_d, 3);
  }
  cout << IOlong;
  if (f & norme) {
    char texte[100]; char *str = norm(dessin->negimage());
    sprintf(texte, "Norme :  %50s", str); free(str);
    textaffiche(10, texte, 50, 30, c_d);
  }
  if (f & valeurs) {
    {
    char texte[100]; char *str = x[0];
    sprintf(texte, "Lambda : %50s", str); free(str);
    textaffiche(11, texte, 50, 50, c_d);
    }
    if (nbq) {
    char texte[100]; char *str = dessin->C();
    sprintf(texte, "C :      %50s", str); free(str);
    textaffiche(12, texte, 50, 70, c_d);
    }
  }
  dessinefenetre();
  if (f & allocation) {
    if (!dessin->dessin_topo)     cerr << "TOPOLOGIE undefined\n";
    if (!dessin->dessin_pos)      cerr << "POSITION undefined\n";
    if (!dessin->dessin_belyi)    cerr << "BELYI undefined\n";
    if (!dessin->dessin_negimage) cerr << "IMAGE undefined\n";
    if (!dessin->dessin_derivee)  cerr << "DERIVEE undefined\n";
  }
}

void
INTERACTIF::loop ()
{
  const size_t stringlength = 100;
  char string[stringlength+1]; string[stringlength] = '\0';
  for (;;) {
    XEVENT e = FENETRE::waitevent();
    switch (e.xev_type) {

    case XEVENT::ignore:
      break;
    case XEVENT::redisplay:
      affiche();
      break;
    case XEVENT::xkeysym: switch (e.xev_ks) {
      case XK_Up:
      case XK_Down:
      case XK_Right:
      case XK_Left:
        winmove (e.xev_ks);
        effacefenetre (); dessinefenetre ();
        affiche (); break;
      } break;
    case XEVENT::caractere: switch (e.xev_char) {

    // **** Fonctions gnrales ******************************************
      // quitter
      case 'q':
        return;
      // aide
      case 'h':
        help ();
        break;
      // effacer et rafficher
      case 'r':
        effacefenetre (); dessinefenetre ();
        affiche (); break;
      // zoom 2
      case '>':
        winscale (2.);
        effacefenetre (); dessinefenetre ();
        affiche (); break;
      // zoom 2
      case '<':
        winscale (.5);
        effacefenetre (); dessinefenetre ();
        affiche (); break;
      // retour vers le pass
      case 'u':
        affiche (ancres, false);
        dessin->undo ();
        affiche (); break;
      // fixe les liens du graphe
      case 'F':
        dessin->fixeliens ();
        affiche (liens); break;
      // demi/quart/huitimes de liens
      case 'X':
        affiche (liens|voisins, false);
        dessin->demiliens ();
        affiche (liens|voisins|points); break;
      // sauvegarde dans un fichier ou affichage
      case 's':
        cout << "Sauver dans le fichier : ";
        cin.get (string,stringlength);
        cin.get (string[stringlength-1]);
        if (strlen (string) == 0)
          cout << *dessin;
        else {
          ofstream file (string);
          if (file) file << *dessin;
          else {
            cerr << "Erreur pour l'criture du fichier " << string;
            perror (" ");
          }
        } break;
      // chargement  partir d'un fichier
      case 'l':
        cout << "Charger  partir du fichier : ";
        cin.get (string,stringlength);
        cin.get (string[stringlength-1]);
        dessin->chargement (string);
        affiche (); break;

    // **** Rsolution du systme ****************************************
      // une itration de Newton
      case ' ':
        try {
          dessin->move ( delta(dessin->position(),
                               dessin->negimage(),
                               dessin->derivee()) );
          affiche ();
        } catch (solve s) {
          switch (s) {
            case nulldet: cerr <<
              "Impossible d'inverser, le dterminant est nul\n"; break;
            case dicho: cerr << "ERROR: Never Reached\n"; break;
          }
        }
        break;
      // itration avec dichotomie
      case '/':
        set_currentDESSIN (dessin);
        try {
          dessin->move ( dichosolve(dessin->position(),
                                    dessin->negimage(),
                                    dessin->derivee(),
                   negimage_currentDESSIN ) );
          affiche ();
        } catch (solve s) {
          switch (s) {
            case nulldet: cerr <<
              "Impossible d'inverser, le dterminant est nul\n"; break;
            case dicho: cerr << "La dichotomie va trop loin\n"; break;
          }
        }
        set_currentDESSIN (NULL);
        break;
      // ancre 0
      case '0':
        affiche (ancres, false);
        dessin->move_ancre (mark, *dessin->ancre0);
        affiche (ancres|norme);
        break;
      // ancre 1
      case '1':
        affiche (ancres, false);
        dessin->move_ancre (mark, *dessin->ancre1);
        affiche (ancres|norme);
        break;
      // ajuste aux ancres
      case '%':
        dessin->ajusteancres ();
        affiche ();
        break;
      // ajuste lambda
      case '&':
        dessin->ajustelambda ();
        affiche (valeurs|norme);
        break;
      // augmente la prcision
      case '+':
        dessin->precision (true);
        break;
      // diminue la prcision
      case '-':
        dessin->precision (false);
        break;
      // valeur algbrique du point marqu
      case 'A':
        polynome_minimal (dessin->position()[mark],
                          dessin->position().length());
        break;

    // **** Modification du dessin ***************************************
      // modification de lambda
      case 'L': {
        cout << "dcalage pour Lambda : ";
        cin.get (string,stringlength);
        cin.get (string[stringlength-1]);
        VECTEUR decalage (dessin->position().length());
        for (ulong i = 0; i < decalage.length(); i++) decalage[i] = 0;
        decalage[0] = string;
        dessin->move (decalage);
        } affiche (valeurs|norme); break;
      // modification de C
      case 'C':
        if (dessin->nombrevaleurs() != 4) {
          cout << "Il n'y a que trois valeurs de ramification\n";
          break;
        } else {
        cout << "Nouvelle valeur de C : ";
        cin.get (string,stringlength);
        cin.get (string[stringlength-1]);
        RE newC = string;
        dessin->moveC (newC);
        } affiche (valeurs|norme); break;
      // modification alatoire de 10%
      case '~': {
        VECTEUR aleatoire (dessin->position().length());
        aleatoire[0] = 0; // On ne modifie pas lambda
        for (ulong i = 1; i < aleatoire.length(); i++) {
          double facteur = drand48() - drand48();
          double mindist = pInf;
          for (ulong j = 0; j < aleatoire.length(); j++) {
            double dist = norm( dessin->position()[i] - dessin->position()[j] );
            if ((dist > 0) && (dist < mindist)) mindist = dist;
          }
          aleatoire[i] = facteur * mindist / 10;
        }
        dessin->move (aleatoire);
        } affiche ();
        break;
      // inversion (marque -> oo)
      case '!':
        affiche (ancres|marque, false);
        if (mark) dessin->inversion (mark);
        mark = 0;
        affiche ();
        break;
      // cration/suppression de la quatrime ramification
      case 'S':
        if (dessin->nombrevaleurs() == 4) {
          affiche (valeurs|ancres|marque, false);
          dessin->unsplit ();
          mark = 0;
          affiche ();
          break;
        }
        if (!mark) {
          cerr << "Erreur: il faut marquer le point  splitter\n";
          break;
        } else {
          TOPOLOGIE::type typemark = dessin->topologie().typeramif(mark);
          if ((typemark!=TOPOLOGIE::sommet) && (typemark!=TOPOLOGIE::arete)) {
            cerr <<
              "Erreur: le point  splitter doit tre un sommet ou une arte\n";
            break;
          }
          if (dessin->topologie().valence(mark) < 2) {
            cerr <<
              "Erreur: le point  splitter doit tre de valence au moins 2\n";
            break;
          }
        }
        // On cherche la liste des voisins
        // si elle a chang, on recommence
        // sinon, on regarde si exactement deux sont slectionns
        // si oui, on splitte, si non, message d'erreur
        SELECTION v = dessin->voisins (mark);
        if (! mark_voisins.same_possible(v)) {
          mark_voisins = v;
          affiche (voisins);
          break;
        }
        if (mark_voisins.nombre() != 2) {
          cerr << "Erreur: il faut slectionner deux voisins\n";
          affiche (liens);
          break;
        }
        affiche (valeurs|ancres|marque, false);
        dessin->split (mark, mark_voisins);
        mark = 0;
        mark_voisins = SELECTION(0,false);
          cout << *dessin;
        affiche ();
        break;

      } break;

    // **** Boutons ******************************************************
    case XEVENT::bouton: switch (e.xev_button) {
      case 1: dessin->info (e.xev_pos); break;
      case 2:
          if (mark_voisins.defini()) { // slection pour split
            dessin->calcule_liens ();
            dessin->demi_positions ();
            const CX& pos_marque = dessin->position()[mark];
            VECTEUR* pos_fin = dessin->liste_voisins (mark);
            __RE_GARBAGE_BEGIN
            double mindist = pInf;
            ulong proche = 100; for (ulong i = 1; i < pos_fin->length(); i++) {
              if (!mark_voisins.is_possible (i)) continue;
              double dist = norm ((*pos_fin)[i] - e.xev_pos);
              if (dist < mindist) { mindist = dist; proche = i; }
            }
            if (double (norm (pos_marque - e.xev_pos)) < mindist)
              mark_voisins = SELECTION(0, false);
            else
              if (mark_voisins.is_possible (proche))
                mark_voisins.swap_possible (proche);
            __RE_GARBAGE_END
            affiche (voisins|liens);
          }
          else { // marque de base
            ulong proche = dessin->voisin (e.xev_pos);
            affiche (marque, false);
            if (proche == mark) mark = 0;
            else                mark = proche;
            affiche (marque);
          }
          dessinefenetre ();
        break;
      case 3: if (mark) {
          affiche (marque, false);
          VECTEUR deplacement (dessin->position().length());
          for (ulong i = 0; i < deplacement.length(); i++) {
            if (i == mark)
              deplacement[i] = e.xev_pos - dessin->position()[mark];
            else
              deplacement[i] = CX (0);
          }
          dessin->move (deplacement);
          affiche (); break;
        } break;
      } break;

    case XEVENT::xevent: 
      cerr << "TODO: XEvent\n";
      break;
    }
  }
  cerr << "ERROR: Never Reached\n";
}

void
help()
{
  cout
<< "  Fonctions gnrales\n"
<< "q : [quit]   quitter             | "
<< "s : [save] sauvegarde dans un fichier\n"
<< "h : [help]   aide                | "
<< "l : [load] chargement  partir d'un fichier\n"
<< "u : [undo] retour vers le pass  | "
<< "r : [redraw] effacer et rafficher\n"
<< "> : zoom 2                      | "
<< "Flches : dplacement du cadre de vue\n"
<< "< : zoom 2                      | "
<< "\n"
<< "F : fixe les liens du graphe     | "
<< "X : demi/quart/huitimes de liens\n"
<< "  Rsolution du systme\n"
<< "  : une itration de Newton      | "
<< "/ : itration avec dichotomie\n"
<< "0 : ancre 0                      | "
<< "% : ajuste aux ancres\n"
<< "1 : ancre 1                      | "
<< "& : ajuste lambda\n"
<< "+ : augmente la prcision        | "
<< "- : diminue la prcision\n"
<< "A : valeur algbrique du point marqu\n"
<< "  Modification du dessin\n"
<< "L : modification de lambda       | "
<< "~ : modification alatoire de 10%\n"
<< "C : modification de C            | "
<< "! : inversion (marque -> oo)\n"
<< "S : [(un)split] cration/suppression de la quatrime ramification\n"
<< "Bouton 1 : informations          | "
<< "Bouton 2 : (d-)marquage du sommet voisin\n"
<< "Bouton 3 : dplacement du sommet marqu\n"
;
}
