#include "polynome.h"

POLYNOME::POLYNOME ()
{
  deg = 0;
  len = 0;
  coef = new CX[len];
}

POLYNOME::POLYNOME (const CX a)
{
  deg = 1;
  len = chunk;
  coef = new CX[len];
  coef[0] = a;
}

POLYNOME::POLYNOME (const POLYNOME& a)
{
  deg = a.deg;
  len = a.len;
  coef = new CX[len];
  for (ulong i = 0; i < deg; i++) coef[i] = a.coef[i];
}

VECTEUR
POLYNOME::vecteur(ulong dim) const
{
  VECTEUR res (dim); ulong howmany = dim;
  if (howmany > deg) howmany = deg;
  ulong i; for (i = 0; i < howmany; i++) res[i] = coef[i];
  for (; i < dim; i++) res[i] = CX(0);
  return res;
}

void
POLYNOME::compute_degre ()
{
  ulong n; for (n = deg; n > 0; n--)
    if (!IsZero(coef[n-1])) break;
  deg = n;
  if ((deg==0)&&(len>0)) deg = 1;
}

void
POLYNOME::operator = (const POLYNOME& a)
{
  // cerr << "POL AFFECT\n";
  this-> ~POLYNOME ();
  deg = a.deg;
  len = a.len;
  coef = new CX[len];
  if (len) for (ulong i = 0; i < deg; i++) coef[i] = a.coef[i];
}

POLYNOME&
POLYNOME::operator += (const POLYNOME& a)
{
  if (!len) return *this;
  ulong i;
  if (a.deg < deg) {
    for (i = 0; i < a.deg; i++) coef[i] += a.coef[i];
    return *this;
  }
  if (a.deg > deg) {
    POLYNOME *r = new POLYNOME(a);
    for (i = 0; i < deg; i++) r->coef[i] += coef[i];
    delete[] coef; coef = r->coef;
    return *this;
  }
  for (i = 0; i < deg; i++) coef[i] += a.coef[i];
  compute_degre();
  return *this;
}

POLYNOME
operator + (const POLYNOME& a, const POLYNOME& b)
{
  ulong i;
  if (a.deg < b.deg) {
    POLYNOME r = b;
    for (i = 0; i < a.deg; i++) r.coef[i] += a.coef[i];
    return r;
  }
  if (a.deg > b.deg) {
    POLYNOME r = a;
    for (i = 0; i < b.deg; i++) r.coef[i] += b.coef[i];
    return r;
  }
  POLYNOME r = a;
  for (i = 0; i < b.deg; i++) r.coef[i] += b.coef[i];
  r.compute_degre();
  return r;
}

POLYNOME
operator - (const POLYNOME& a, const POLYNOME& b)
{
  ulong i;
  if (a.deg < b.deg) {
    POLYNOME r = b;
    for (i = 0; i < a.deg; i++) r.coef[i] -= a.coef[i];
    return r;
  }
  if (a.deg > b.deg) {
    POLYNOME r = a;
    for (i = 0; i < b.deg; i++) r.coef[i] -= b.coef[i];
    return r;
  }
  POLYNOME r = a;
  for (i = 0; i < b.deg; i++) r.coef[i] -= b.coef[i];
  r.compute_degre();
  return r;
}

POLYNOME&
POLYNOME::operator *= (const CX& k)
{
  ulong i;
  for (i = 0; i < deg; i++) coef[i] *= k;
  if (IsZero(k)&&(len>0)) deg = 1;
  return *this;
}

POLYNOME
operator * (const CX& k, const POLYNOME& a)
{
  POLYNOME r = a; r *= k; return r;
}

// Multiplication par (X-a)^n
void
POLYNOME::mul_xan (const CX& a, ulong n)
{
  if (len==0) return;
  if (deg + n > len) {
    ulong newlength = deg + n - 1;
    newlength /= chunk; newlength++; newlength *= chunk;
    CX* newcoef = new CX[newlength];
    for (ulong i = 0; i < deg; i++) newcoef[i] = coef[i];
    delete[] coef; coef = newcoef; len = newlength;
  }
  CX ma = -a;
  for ( ; n; n--) {
    coef[deg++] = CX(0);
    for (ulong i = deg-1; i; i--) { coef[i] *= ma; coef[i] += coef[i-1]; }
    coef[0] *= ma;
  }
}

// Division par (X-a)^n
void
POLYNOME::div_xan (const CX& a, ulong n)
{
  if (len==0) return;
  if (deg==1) {
    cerr << "On ne peut diviser un polynme constant...\n";
    return;
  }
  // cout << "diviser " << *this << " par X-" << a << "\n";
  for ( ; n; n--) {
    for (ulong i = deg-1; i; i--) coef[i-1] += a * coef[i];
    for (ulong i = 0; i < deg-1; i++) coef[i] = coef[i+1];
    coef[--deg] = CX(0);
    if (deg==0) cout << "erreur\n";
  }
}

void
POLYNOME::derivee ()
{
  for (ulong i = 0; i < deg-1; i++)
    coef[i] = double(i+1) * coef[i+1];
  coef[--deg] = CX(0);
  if (deg==0) deg++;
}

ostream&
operator << (ostream& s, const POLYNOME& a)
{
  ulong i;
  for (i = a.degre(); i; i--) {
    s << "(" << a[i] << ")*X^" << i << " + ";
  } s << a[0];
  return s;
}
