#include "vecteur.h"

// Constructeurs et destructeurs par dfaut
VECTEUR::~VECTEUR ()
{
  (*refcount) --;
  if (!*refcount) { delete[] val; delete refcount; }
}

MATRICE::~MATRICE ()
{
  (*refcount) --;
  if (!*refcount) { delete[] val; delete refcount; }
}

VECTEUR::VECTEUR (ulong s)
{
  dim = s;
  val = new CX[dim];
  refcount = new ulong (1);
}

MATRICE::MATRICE (ulong l, ulong c)
{
  nbl = l;
  nbc = c;
  val = new CX[l*c];
  refcount = new ulong (1);
}


// Copie de vecteur et de matrice
VECTEUR::VECTEUR (const VECTEUR& a)
{
  // cout << "VEC COPY\n";
  dim = a.dim;
  val = a.val; 
  refcount = a.refcount;
  (*refcount) ++;
}

MATRICE::MATRICE (const MATRICE& a)
{
  // cout << "MAT COPY\n";
  nbl = a.nbl;
  nbc = a.nbc;
  val = a.val; 
  refcount = a.refcount;
  (*refcount) ++;
}

void
VECTEUR::copy_if_multiref ()
{
  if (*refcount == 1) return;
  CX * oldval = val; val = new CX[dim];
  ulong i; for (i = 0; i < dim; i++) val[i] = oldval[i];
  (*refcount) --; refcount = new ulong (1);
}

void
MATRICE::copy_if_multiref ()
{
  if (*refcount == 1) return;
  CX * oldval = val; val = new CX[nbl*nbc];
  ulong i; for (i = 0; i < nbl*nbc; i++) val[i] = oldval[i];
  (*refcount) --; refcount = new ulong (1);
}

void
VECTEUR::operator = (const VECTEUR& a)
{
  this-> ~VECTEUR();
  dim = a.dim;
  val = a.val; 
  refcount = a.refcount;
  (*refcount) ++;
}

void
MATRICE::operator = (const MATRICE& a)
{
  this-> ~MATRICE();
  nbl = a.nbl;
  nbc = a.nbc;
  val = a.val; 
  refcount = a.refcount;
  (*refcount) ++;
}


// Oprations sur les vecteurs
VECTEUR&
VECTEUR::operator += (const CX& c)
{
  CX * oldval = val;
  val = new CX[dim+1];
  ulong i; for (i = 0; i < dim; i++) val[i] = oldval[i];
  val[dim] = c; dim++;
  (*refcount) --;
  if (!*refcount) { delete[] oldval; *refcount = 1; }
  else        { refcount = new ulong (1); }
  return *this;
}

VECTEUR&
VECTEUR::operator *= (const CX& c)
{
  ulong i;
  if (*refcount > 1) {
    CX * oldval = val;
    val = new CX[dim];
    for (i = 0; i < dim; i++) val[i] = oldval[i] * c;
    (*refcount) --; refcount = new ulong (1);
  }
  else
    for (i = 0; i < dim; i++) val[i] *= c;
  return *this;
}

VECTEUR&
VECTEUR::operator /= (const CX& c)
{
  ulong i;
  if (*refcount > 1) {
    CX * oldval = val;
    val = new CX[dim];
    for (i = 0; i < dim; i++) val[i] = oldval[i] / c;
    (*refcount) --; refcount = new ulong (1);
  }
  else
    for (i = 0; i < dim; i++) val[i] /= c;
  return *this;
}

VECTEUR&
VECTEUR::operator += (const VECTEUR& a)
{
  if ( dim != a.dim ) {
    cerr << "Error: can't add vectors with different length\n";
    return *this;
  }
  ulong i;
  if (*refcount > 1) {
    CX * oldval = val;
    val = new CX[dim];
    for (i = 0; i < dim; i++) val[i] = oldval[i] + a.val[i];
    (*refcount) --; refcount = new ulong (1);
  }
  else
    for (i = 0; i < dim; i++) val[i] += a.val[i];
  return *this;
}

VECTEUR&
VECTEUR::operator -= (const VECTEUR& a)
{
  if ( dim != a.dim ) {
    cerr << "Error: can't add vectors with different length\n";
    return *this;
  }
  ulong i;
  if (*refcount > 1) {
    CX * oldval = val;
    val = new CX[dim];
    for (i = 0; i < dim; i++) val[i] = oldval[i] - a.val[i];
    (*refcount) --; refcount = new ulong (1);
  }
  else
    for (i = 0; i < dim; i++) val[i] -= a.val[i];
  return *this;
}

RE
norm (const VECTEUR& a)
{
  RE x (0);
  ulong i; for (i = 0; i < a.length(); i++) x += norm(a[i]);
  return x;
}

// Affichage
ostream&
operator << (ostream& s, const VECTEUR& a)
{
  s << "["; ulong i; for (i = 0; i+1 < a.length(); i++) s << a[i] << " ";
  if (a.length()) s << a[a.length()-1]; s << "]\n";
  return s;
}

ostream&
operator << (ostream& s, const MATRICE& a)
{
  ulong j; for (j = 0; j < a.lig(); j++) {
    s << "["; ulong i; for (i = 0; i+1 < a.col(); i++) s << a(j,i) << " ";
    if (a.col()) s << a(j,a.col()-1); s << "]\n";
  }
  return s;
}

// Manipulation de matrice
VECTEUR
MATRICE::operator * (const VECTEUR& x) const
{
  ulong i, j;
  VECTEUR y(nbl);
  for (i = 0; i < nbl; i++) y[i] = CX(0);
  if (nbc == x.length()) {
    for (i = 0; i < nbc; i++) for (j = 0; j < nbc; j++)
      y[i] += x[j] * val[j+i*nbc];
  } else {
    cerr << "MAT VEC mul : col " << nbc << " != len " << x.length() << "\n";
  }
  return y;
}

CX
matinv (const MCARREE& matrice, MCARREE& inverse)
{
  __RE_GARBAGE_BEGIN
  ulong d = matrice.dim();
  ulong lig, col;
#ifdef __GNUC__
  CX **mi = new (CX*)[d];
#else
  CX **mi = (CX**) malloc(sizeof(CX*)*d);
#endif
  for (lig = 0; lig < d; lig++) {
    mi[lig] = new CX[2*d];
    for (col = 0; col < d; col++) mi[lig][col] = matrice(lig,col);
    for (col = 0; col < d; col++) mi[lig][col+d] = CX(col==lig?1:0);
  }
  CX det = 1;
  ulong pos, k; double maxval;
  for (k = 0; k < d; k++) {
    // Recherche du pivot
    pos = d; maxval = 0.;
    for (lig = k; lig < d; lig++) {
      double t2 = norm (mi[lig][k]);
      if (t2 > maxval) { pos = lig; maxval = t2; }
    }
    if (pos != d) {
    // change et soustraction de la ligne pivot
      if (pos == k) det *= mi[k][k];
      else { CX* t2=mi[k]; mi[k]=mi[pos]; mi[pos]=t2; det *= - mi[k][k]; }
      // make mi[k, k] == -1
      CX inversevalue = - inv(mi[k][k]);
      for (col = k+1; col < 2*d; col++) mi[k][col] *= inversevalue;
      // mi[k][k] = -1; // inutile
      for (lig = k+1; lig < d; lig++) {
        for (col = k+1; col < 2*d; col++)
          mi[lig][col] += mi[k][col] * mi[lig][k];
        // mi[lig][k] = 0; // inutile
      }
    }
    else { det = 0; k = d + 1; }
  }
  if (k <= d) {
    for (k = 0; k < d; k++) for (lig = d; lig; lig--) {
      CX val (0);
      for (col = lig; col < d; col++)
        val += inverse(col, k) * mi[lig-1][col];
      val -= mi[lig-1][k+d];
      inverse(lig-1, k) = val;
    }
  }
  for (lig = 0; lig < d; lig++) delete[] mi[lig];
#ifdef __GNUC__
  delete[] mi;
#else
  free(mi);
#endif
#if defined(mplib_pari)
  GEN *gptr [2*d*d+2];
  {for (ulong lig = 0; lig < d; lig++)
    for (ulong col = 0; col < d; col++)
      genptr (gptr, col+lig*d, inverse(lig, col));}
  genptr (gptr, d*d, det);
  gerepilemany(avma_garbage, gptr, 2*d*d+2);
#endif
  return det;
}
